require("defines")
-- debug_status = 1
debug_file = "dron-deb.txt"
debug_mod_name = "Drones"
require("utils")
require("config")

local nb_drones_per_page_init = 10

--------------------------------------------------------------------------------------
function on_init()	
	global.ticks = global.ticks or 0
	global.players_mem = global.players_mem or {}
	global.surface_hidden = global.surface_hidden or nil
	global.antennas = global.antennas or {}
	global.refresh = global.refresh or false
end

script.on_init( on_init )

--------------------------------------------------------------------------------------
function on_configuration_changed(data)
	if data.mod_changes ~= nil and data.mod_changes["Drones"] ~= nil then
		debug_print( "update mod: Drones" )
		
		on_init()
		
		local force
		for _,force in pairs(game.forces) do
			force.reset_recipes()
			force.reset_technologies()
			force.recipes["drone-antenna"].enabled = force.technologies["drones-surveillance"].researched
		end
		
		if data.mod_changes["CameraDrones"] ~= nil and data.mod_changes["CameraDrones"] ~= nil and data.mod_changes["CameraDrones"].old_version <= "1.0.3" then
			debug_print( "migrate mod: CameraDrones" )
			local chunk, ent, surf, a, nb_surv, nb_maint
			nb_surv=0
			nb_maint=0
			surf = game.surfaces["nauvis"]
			for chunk in surf.get_chunks() do
				a = {{chunk.x * 32, chunk.y * 32}, {chunk.x * 32 + 32, chunk.y * 32 + 32}}
				for _, ent in pairs(surf.find_entities_filtered({area=a, name= "camera-drone"})) do
					ent.destroy()
					nb_surv=nb_surv+1
				end
				for _, ent in pairs(surf.find_entities_filtered({area=a, name= "drone-surveillance"})) do
					ent.destroy()
					nb_surv=nb_surv+1
				end
				for _, ent in pairs(surf.find_entities_filtered({area=a, name= "camera-drone-maint"})) do
					ent.destroy()
					nb_maint=nb_maint+1
				end
				for _, ent in pairs(surf.find_entities_filtered({area=a, name= "drone-maintenance"})) do
					ent.destroy()
					nb_maint=nb_maint+1
				end
			end
			
			local player
			for _,player in pairs(game.players) do
				player.print( "Give back " .. nb_surv .. "+" .. nb_maint .. " drones." )
				if nb_surv > 0 then player.insert( {name = "drone-surveillance", count = nb_surv} ) end
				if nb_maint > 0 then player.insert( {name = "drone-maintenance", count = nb_maint} ) end
				if nb_surv+nb_maint > 0 then player.force.technologies["drones-surveillance"].researched = true end
			end	
		end

		-- if global.surface_hidden == nil then
			-- global.surface_hidden = game.create_surface("drones-hidden", {width = 1, height = 1, peaceful_mode = true})
		-- end

		local player
		for _,player in pairs(game.players) do
			init_player( player )
		end	
	end
end

script.on_configuration_changed(on_configuration_changed)

--------------------------------------------------------------------------------------
function on_research_finished(event)
	if event.research ~= nil and (event.research.name == "drones-surveillance" or event.research.name == "drones-maintenance") then
		local player
		for _,player in pairs(game.players) do
			init_player( player, true )
		end
	end
end

script.on_event(defines.events.on_research_finished, on_research_finished )

--------------------------------------------------------------------------------------
function on_creation( event )
	local ent = event.created_entity
	
	if ent.name == "drone-surveillance" or ent.name == "drone-maintenance" then
		local player = game.get_player(event.player_index)
		
		if player.character.name == "player" then
			add_drone( player, ent )
		else
			player.print({"gui-drones-cannot-drone"})
			ent.destroy()
		end
	elseif ent.name == "drone-antenna" then
		table.insert( global.antennas, ent )
		global.refresh = true
	end
end

script.on_event(defines.events.on_built_entity, on_creation )
script.on_event(defines.events.on_robot_built_entity, on_creation )

--------------------------------------------------------------------------------------
function on_destruction( event )
	local ent = event.entity

	if ent.name == "drone-antenna" then
		local k, ant
		for k, ant in pairs(global.antennas) do
			if ant == ent then 
				table.remove(global.antennas, k)
				break
			end
		end
		global.refresh = true
	end
end

script.on_event(defines.events.on_entity_died, on_destruction )
script.on_event(defines.events.on_robot_pre_mined, on_destruction )
script.on_event(defines.events.on_preplayer_mined_item, on_destruction )

--------------------------------------------------------------------------------------
function on_player_created(event)
	debug_print( "player created" )
	local player = game.get_player(event.player_index)
	init_player( player )
	
	if debug_status == 1 or is_dev(player) then
		player.insert({name = "drone-surveillance", count = 10} )
		player.insert({name = "drone-maintenance", count = 5} )
		player.insert({name = "drone-antenna", count = 5} )
	end
end

script.on_event(defines.events.on_player_created, on_player_created )

--------------------------------------------------------------------------------------
function on_tick(event)
	if global.ticks <= 0 then
		global.ticks = 124
		--global.ticks = 500
	else
		global.ticks = global.ticks - 1
	end
	
	if global.ticks == 119 then
		local player, player_mem, drone
		for _,player in pairs(game.players) do
			if player.connected then
				player_mem = global.players_mem[player.index]
				if player_mem and player_mem.drone_cur then -- in drone
					if not (player_mem.character.valid and player_mem.character.name == "player") then -- test if body dead while in drone
						player.print( {"gui-drones-body-died"} )
						-- create new body to sacrifice
						local character = player.surface.create_entity({name = "player", position = player.position, force = player.force})
						player.character = character -- reintegrate new body
						player_mem.character = character
						player_mem.drone_cur.entity.active = false -- in case moving when body die
						player_mem.drone_cur = nil
						player_mem.drone_cur_cover = -1
						player.character.die() -- kill new body
					end
				end
			end
		end
		
	elseif global.ticks == 78 then
		local player, player_mem
		for _,player in pairs(game.players) do
			if player.connected then
				-- test if del confirm out of time
				player_mem = global.players_mem[player.index]
				if player_mem.drone_del_confirm then
					if game.tick > player_mem.drone_del_confirm_tick then
						player_mem.drone_del_confirm = nil
						if player.gui.left.flow_drones then update_menu_drones(player) end
					end
				end
				-- test if refresh needed
				if global.refresh and player.gui.left.flow_drones then 
					update_menu_drones(player)
					global.refresh = false
				end
			end
		end
		
	elseif global.ticks == 37 then
		local player, player_mem
		for _,player in pairs(game.players) do
			if player.connected and player.character then
				-- test if item in drone slots, and put it back in toolbelt
				if player.character.name == "drone-surveillance" then
					check_empty_inventory(player,defines.inventory.player_tools)
					check_empty_inventory(player,defines.inventory.player_guns)
					check_empty_inventory(player,defines.inventory.player_ammo)
					check_empty_inventory(player,defines.inventory.player_armor)
				elseif player.character.name == "drone-maintenance" then
					check_empty_inventory(player,defines.inventory.player_guns)
					check_empty_inventory(player,defines.inventory.player_ammo)
					check_empty_inventory(player,defines.inventory.player_armor)
				end
			end
		end
		
	elseif global.ticks%30 == 0 then
		local player, player_mem
		for _,player in pairs(game.players) do
			if player.connected then
				-- test if drone in antenna range
				check_antenna_range(player)
			end
		end
		
	end
end

script.on_event(defines.events.on_tick, on_tick)

--------------------------------------------------------------------------------------
function check_antenna_range(player)
	local player_mem = global.players_mem[player.index]
	
	local cover, drone
	local changes = false
	
	for _, drone in pairs( player_mem.drones ) do
		if calc_antenna_range(drone) then
			changes = true
			cover = drone.cover
			
			if drone == player_mem.drone_cur and cover ~= player_mem.drone_cur_cover then
				player_mem.drone_cur_cover = cover

				if cover == 2 then
					player_mem.drone_cur.entity.active = true
					player.surface.create_entity({name = "flying-text", position = pos_offset(player.position,{x=-0.5,y=0.2}), text = {"gui-drones-flying-ok"}, color = {r=0, g=1, b=0}})
				elseif cover == 1 then
					player.print( {"gui-drones-antenna-warning"} )
					player.surface.create_entity({name = "flying-text", position = pos_offset(player.position,{x=-0.5,y=0.2}), text = {"gui-drones-flying-warning"}, color = {r=1, g=1, b=0}})
				else
					player.print( {"gui-drones-antenna-lost"} )
					player_mem.drone_cur.entity.active = false
					switch_to_character( player, false )
					player_mem.drone_dup = nil
					player_mem.drone_cur = nil
					player.surface.create_entity({name = "flying-text", position = pos_offset(player.position,{x=-0.5,y=0.2}), text = {"gui-drones-flying-lost"}, color = {r=1, g=0, b=0}})
				end
			end
		end
	end	
	
	if changes then
		update_menu_drones(player)				
	end
	
	if player_mem.drone_cur and player_mem.gui_dist then
		player_mem.gui_dist.caption = "@" .. math.floor(player_mem.drone_cur.dist) .. "m"
	end
end

--------------------------------------------------------------------------------------
function calc_antenna_range(drone)
	-- return true if cover changed

	if drone == nil or drone.entity == nil or not drone.entity.valid then 
		drone.cover = 2
		return(true) 
	end
	
	local result = 0
	local pos = drone.entity.position
	local force = drone.entity.force
	local ant, dist
	local distMin = 0
	local result = 0
	
	for _, ant in pairs(global.antennas) do
		if ant.force == force and ant.energy > 0 then
			dist = distance_square( ant.position, pos )
			if distMin == 0 then
				distMin = dist
			else
				distMin = min(distMin,dist)
			end
			if dist <= drone_antenna_radius_warning then
				result = 2
				break
			elseif dist <= drone_antenna_radius_offline then
				result = 1
			end
		end
	end
	
	drone.dist = distMin
	
	if drone.cover == result then return( false ) end
	
	drone.cover = result
	return( true )
end		

--------------------------------------------------------------------------------------
function on_gui_click(event)
	-- debug_print( "player index=", event.player_index, " event=", event.element.name )
	
	if event.element.name == "but_drones" then
		-- main menu toggle
		local player = game.get_player(event.player_index)
		local player_mem = global.players_mem[player.index]
		
		if player.gui.left.flow_drones then
			player.gui.left.flow_drones.destroy()
			--player_mem.gui_opened = false
			player_mem.gui_dist = nil
		else
			update_menu_drones(player)
			--player_mem.gui_opened = true
		end
		
	elseif event.element.name == "but_drone_pagedown" then
		local player = game.get_player(event.player_index)		
		local player_mem = global.players_mem[player.index]
		
		player_mem.pagecur = max( 1, player_mem.pagecur - 1 )
		update_menu_drones(player)
		
	elseif event.element.name == "but_drone_pageup" then
		local player = game.get_player(event.player_index)		
		local player_mem = global.players_mem[player.index]
		
		player_mem.pagecur = min( player_mem.pagenb, player_mem.pagecur + 1 )
		update_menu_drones(player)

	elseif event.element.name == "but_drone_body" then
		-- switch to man
		local player = game.get_player(event.player_index)		
		local player_mem = global.players_mem[player.index]

		if player.walking_state.walking then
			player.print( {"gui-drones-stop-moving"} )
		else
			-- back to character
			if player_mem.drone_cur and not player_mem.drone_dup then 
				switch_to_character( player, true )
				player_mem.drone_cur = nil
				player_mem.drone_cur_cover = -1
				update_menu_drones(player)
			end
		end

	elseif event.element.name == "but_drone_dup" then
		local player = game.get_player(event.player_index)	

		if player.walking_state.walking then
			player.print( {"gui-drones-stop-moving"} )
		else
			local player_mem = global.players_mem[player.index]
			
			if player.character and player.character.valid and player.character.name == "player" then -- only if in body
				-- dup
				local drone = player.surface.create_entity({ name = "drone-surveillance", force = player.force, position = pos_offset(player.position,{x=0,y=-1}) })
				local drone_new = add_drone(player, drone)
				
				if drone_new then
					if switch_to_drone( player, player_mem, drone_new ) then
						player_mem.drone_dup = drone_new
					end
					
					update_menu_drones(player)
				end
			else
				if player_mem.drone_dup then
					-- back and del dup
					local drone,i
					
					for i, drone in pairs(player_mem.drones) do
						if drone == player_mem.drone_dup then
							table.remove(player_mem.drones,i)
							break
						end
					end
					
					switch_to_character( player, true )
					player.teleport(player_mem.drone_dup.entity.position, player_mem.drone_dup.entity.surface)
					player_mem.drone_dup.entity.destroy()
					player_mem.drone_dup = nil
					player_mem.drone_cur = nil
					player_mem.drone_cur_cover = -1
					update_menu_drones(player)				
				end
			end
		end
		
	elseif event.element.name == "but_drone_port" then
		local player = game.get_player(event.player_index)	
		local player_mem = global.players_mem[player.index]
		
		if player_mem.drone_cur then -- only if in drone
			local character = player_mem.character
			if character and character.valid then 
				switch_to_character( player, true )
				player.teleport(player_mem.drone_cur.entity.position,player_mem.drone_cur.entity.surface)
				player_mem.drone_cur = nil
				player_mem.drone_cur_cover = -1
				update_menu_drones(player)
			end
		end
		
	elseif string.sub(event.element.name, 1, 14) == "but_drone_del_" then
		-- delete drone
		local player = game.get_player(event.player_index)		
		local player_mem = global.players_mem[player.index]
		local n_drone = tonumber(string.sub(event.element.name, 15))
		local drone_del = player_mem.drones[n_drone]
		
		if drone_del then
			if player.character.name ~= "player" then
				player.print({"gui-drones-cannot-delete-inside"})
			elseif distance_square( player.position, drone_del.entity.position ) > drone_recover_radius then
				player.print({"gui-drones-cannot-delete-far"})
			elseif player_mem.drone_del_confirm == drone_del then
				local drone_name = drone_del.entity.name
				
				player_mem.drone_cur = nil
				player_mem.drone_cur_cover = -1
				player_mem.drone_dup = nil
				
				if prepare_del_drone(player,drone_del.entity) then
					if player.can_insert(drone_name) then
						local drone,i
						
						for i, drone in pairs(player_mem.drones) do
							if drone == drone_del then
								table.remove(player_mem.drones,i)
								break
							end
						end
						
						drone_del.entity.destroy()
						player.insert({name=drone_name, count=1})
					else
						player.print( {"gui-drones-cannot-delete-space"} )
					end
				else
					player.print( {"gui-drones-cannot-delete-space"} )
				end
			else
				player.print( {"gui-drones-confirm-delete"} )
				player_mem.drone_del_confirm = drone_del
				player_mem.drone_del_confirm_tick = game.tick + 180
			end
			
			update_menu_drones(player)
		end
		
	elseif string.sub(event.element.name, 1, 14) == "but_drone_ren_" then
		-- rename drone
		local player = game.get_player(event.player_index)		
		local player_mem = global.players_mem[player.index]
		local n_drone = tonumber(string.sub(event.element.name, 15))
		local drone_ren = player_mem.drones[n_drone]
		
		if drone_ren then
			player_mem.drone_ren = drone_ren
			show_rename(player, drone_ren)
			
			update_menu_drones(player)
		end
		
	elseif string.sub(event.element.name, 1, 14) == "but_drone_swi_" then
		-- switch from man to drones
		local player = game.get_player(event.player_index)		
		local player_mem = global.players_mem[player.index]
		local n_drone = tonumber(string.sub(event.element.name, 15))
		local drone_new = player_mem.drones[n_drone]
		calc_antenna_range(drone_new)

		if player.walking_state.walking then
			player.print( {"gui-drones-stop-moving"} )
		elseif drone_new.cover == 0 then
			player.print( {"gui-drones-antenna-offline"} )
		else
			if drone_new and not player_mem.drone_dup then
				local drone_cur = player_mem.drone_cur
				
				if drone_new and drone_new ~= drone_cur then
					-- to drone
					if switch_to_drone( player, player_mem, drone_new ) then
						player_mem.drone_cur = drone_new
					end
				else
					-- back to character
					if drone_cur then 
						switch_to_character( player, true )
						player_mem.drone_cur = nil
						player_mem.drone_cur_cover = -1
						player_mem.drone_dup = nil
					end
				end
				
				update_menu_drones(player)
			end
		end

	elseif string.sub(event.element.name, 1, 14) == "but_drone_ptr_" then
		-- point to drone
		local player = game.get_player(event.player_index)		
		local player_mem = global.players_mem[player.index]
		local n_drone = tonumber(string.sub(event.element.name, 15))
		local drone = player_mem.drones[n_drone]
		
		if drone and drone.entity.valid then
			if player_mem.pointer == true then
				player.clear_gui_arrow()
				player_mem.pointer = false
			else
				player.set_gui_arrow( { type = "entity", entity = drone.entity } )
				player_mem.pointer = true
			end
		end

	elseif event.element.name == "but_drone_rename_ok" then
		local player = game.get_player(event.player_index)	
		local player_mem = global.players_mem[player.index]
		
		if player.gui.center.frame_drone_rename ~= nil then
			local s = player.gui.center.frame_drone_rename.textfield_drone_rename.text
			if s == "" then s = "???" end
			player_mem.drone_ren.name = s
			player.gui.center.frame_drone_rename.destroy()
			update_menu_drones(player)
		end
		
		player_mem.drone_ren = nil
		
	elseif event.element.name == "but_drone_rename_cancel" then
		local player = game.get_player(event.player_index)	
		local player_mem = global.players_mem[player.index]
		
		if player.gui.center.frame_drone_rename ~= nil then
			player.gui.center.frame_drone_rename.destroy()
		end
		
		player_mem.drone_ren = nil
	end
end

script.on_event(defines.events.on_gui_click, on_gui_click )

--------------------------------------------------------------------------------------
function init_player( player, techno_ok )
	if player.gui.top.but_drones then player.gui.top.but_drones.destroy() end
		
	if techno_ok == true or player.force.technologies["drones-surveillance"].researched then
		player.gui.top.add({type = "button", name = "but_drones", style = "drones_main_button_style"})	
	end
	
	-- original variables
	
	local players_mem =
	{ 
		character = nil, -- record player main character
		gui_opened = false,
		pagecur = 1, -- current page (1 .. pagenb)
		pagenb = 1, -- total number of pages
		nb_drones_per_page = nb_drones_per_page_init,
		drone_cur = nil, -- current drone if droning
		drone_ren = nil, -- drone being renamed
		drones = {}, -- list of drones = {entity, name, is_maint, cover, dist}
		force = player.force, 
		color = player.color,
	}
	
	if player.connected then players_mem.character = player.character end
	
	global.players_mem[player.index] = global.players_mem[player.index] or players_mem

	-- additionnal late variables 
	players_mem = global.players_mem[player.index]

	players_mem.tick_last_creation = players_mem.tick_last_creation or 0
	players_mem.drone_del_confirm = players_mem.drone_del_confirm or nil
	players_mem.drone_del_confirm_tick = players_mem.drone_del_confirm_tick or 0
	players_mem.drone_dup = players_mem.drone_dup or nil
	players_mem.no_drone = players_mem.no_drone or 0
	players_mem.pointer = players_mem.pointer or false
	players_mem.drone_cur_cover = players_mem.drone_cur_cover or -1
	players_mem.gui_dist = players_mem.gui_dist or nil
end

--------------------------------------------------------------------------------------
function check_empty_inventory( player, inv_type )
	local inv_weapon = player.get_inventory(inv_type)
	if not inv_weapon.is_empty() then
		player.print( {"gui-drones-not-wear-this"} )
		local stack, i
		for i = 1, #inv_weapon do
			stack = inv_weapon[i]
			if stack.valid_for_read then 
				local cursor = player.cursor_stack
				if cursor and cursor.valid_for_read then
					player.get_inventory(defines.inventory.player_quickbar).insert(stack)
				else
					player.cursor_stack.set_stack(stack)
				end
				inv_weapon.remove(stack)
			end
		end
	end
end

--------------------------------------------------------------------------------------
function add_drone( player, ent )
	local player_mem = global.players_mem[player.index]
	
	if game.tick > player_mem.tick_last_creation + 15 then
		-- if ent.name == "drone-surveillance" then ent.destructible = false end
		ent.destructible = false
		
		if player ~= nil and #player_mem.drones < 100 then
			local drone_new
			player_mem.no_drone = player_mem.no_drone + 1
			if ent.name == "drone-maintenance" then
				drone_new = { entity = ent, name = "Drone #" .. player_mem.no_drone, is_maint = true, cover = -1, dist = 0 }
				--drone_new = { entity = ent, name = "Drone #" .. player_mem.no_drone }
			else
				drone_new = { entity = ent, name = "Drone #" .. player_mem.no_drone, is_maint = false, cover = -1, dist = 0 }
				--drone_new = { entity = ent, name = "Drone #" .. player_mem.no_drone }
			end
			table.insert( player_mem.drones, drone_new )
			player_mem.tick_last_creation = game.tick
			update_menu_drones(player)
			return(drone_new)
		end
	else
		player.insert( {name = ent.name, count = 1 } )
		ent.destroy()
	end
	
	return( nil )
end

--------------------------------------------------------------------------------------
function prepare_del_drone( player, ent )
	local result = true
	result = result and move_inventory( player, ent.get_inventory(defines.inventory.player_main) )
	result = result and move_inventory( player, ent.get_inventory(defines.inventory.player_quickbar) )
	result = result and move_inventory( player, ent.get_inventory(defines.inventory.player_guns) )
	result = result and move_inventory( player, ent.get_inventory(defines.inventory.player_ammo) )
	result = result and move_inventory( player, ent.get_inventory(defines.inventory.player_armor) )
	result = result and move_inventory( player, ent.get_inventory(defines.inventory.player_tools) )
	result = result and move_inventory( player, ent.get_inventory(defines.inventory.player_trash) )
	return(result)
end

--------------------------------------------------------------------------------------
function move_inventory( player, inv )
	local stack, i, result, count, name, nb
	local ok = true
	local contents = inv.get_contents()
	
	for name, nb in pairs(contents) do
		player.print( name .. "=" .. nb )
		result = player.insert({name = name, count = nb})
		if result > 0 then
			inv.remove({name = name, count = result})
		end
		if result < nb then
			ok = false
		end
	end
	
	return( ok )
end

--------------------------------------------------------------------------------------
function update_menu_drones(player)
	local guif = player.gui.left.flow_drones
	local player_mem = global.players_mem[player.index]
	
	if player.gui.left.flow_drones then
		player.gui.left.flow_drones.destroy()
	end
	
	if #player_mem.drones > 0 then
		local i, drone, guif2, but

		player_mem.pagenb = 1 + math.floor( (#player_mem.drones-1) / player_mem.nb_drones_per_page )
		
		guif = player.gui.left.add({type = "flow", name = "flow_drones", direction = "vertical", style = "drones_flow_style"})
		guif2 = guif.add({type = "frame", name = "flow_body", direction = "horizontal", style = "drones_frame_style"})
		guif2.add({type = "button", name = "but_drone_pagedown", caption = "<", style = "drones_button_style"})
		guif2.add({type = "button", name = "but_drone_pagecur", caption = player_mem.pagecur .. "/" .. player_mem.pagenb, style = "drones_button_style"})
		guif2.add({type = "button", name = "but_drone_pageup", caption = ">", style = "drones_button_style"})
		but = guif2.add({type = "button", name = "but_drone_body", caption = {"gui-drones-body"}, style = "drones_button_style"})
		if player_mem.drone_cur == nil then but.style.font_color = colors.green end
		
		if debug_status == 1 or is_dev(player) then
			but = guif2.add({type = "button", name = "but_drone_dup", caption = "Dup", style = "drones_button_style"})
			if player_mem.drone_dup then but.style.font_color = colors.yellow end
			guif2.add({type = "button", name = "but_drone_port", caption = "Port", style = "drones_button_style"})
		end
		
		player_mem.gui_dist = nil
		
		for i, drone in pairs(player_mem.drones) do
			if i > player_mem.pagecur * player_mem.nb_drones_per_page then break end
			if i >= 1 + (player_mem.pagecur-1) * player_mem.nb_drones_per_page then
				guif2 = guif.add({type = "frame", name = "flow_drone_" .. i, direction = "horizontal", style = "drones_frame_style"})
				but = guif2.add({type = "button", name = "but_drone_del_" .. i, caption = "X", style = "drones_button_style"})
				if drone == player_mem.drone_del_confirm then but.style.font_color = colors.red end
				but = guif2.add({type = "button", name = "but_drone_ren_" .. i, caption = "N", style = "drones_button_style"})
				if drone == player_mem.drone_del_confirm then but.style.font_color = colors.red end
				but = guif2.add({type = "button", name = "but_drone_ptr_" .. i, caption = "?", style = "drones_button_style"})
				but = guif2.add({type = "button", name = "but_drone_swi_" .. i, caption = drone.name, style = "drones_button_style"})
				if drone.is_maint == true then but.style.font_color = colors.orange end
				if drone == player_mem.drone_cur then but.style.font_color = colors.green end
				if drone == player_mem.drone_dup then but.style.font_color = colors.yellow end
				if drone == player_mem.drone_del_confirm then but.style.font_color = colors.red end
				but = guif2.add({type = "label", name = "lab_drone_elc_" .. i, caption = "@" .. math.floor(drone.dist) .. "m", style = "drones_label_style"})
				if drone.cover == 2 then
					-- but.caption = "hi"
					but.style.font_color = colors.green
				elseif drone.cover == 1 then
					-- but.caption = "lo"
					but.style.font_color = colors.yellow
				else
					-- but.caption = "!!"
					but.style.font_color = colors.red
				end
				if drone == player_mem.drone_cur then player_mem.gui_dist = but end
			end
		end
		
		--player_mem.gui_opened = true
	else
		--player_mem.gui_opened = false
	end
	
	return(guif)
end

--------------------------------------------------------------------------------------
function show_rename(player,drone)
	local guif = player.gui.center.frame_drone_rename
	local player_mem = global.players_mem[player.index]
	
	if guif == nil then
		guif = player.gui.center.add({type = "frame", name = "frame_drone_rename", caption = {"gui-drones-rename"}, direction = "horizontal", style = "drones_frame_style"})
		guif.add({type = "textfield", name = "textfield_drone_rename", text = " ", style = "drones_textfield_style"})
		guif.add({type = "button", name = "but_drone_rename_ok", caption = "OK", style = "drones_button_style"})
		guif.add({type = "button", name = "but_drone_rename_cancel", caption = {"gui-drones-rename-cancel"}, style = "drones_button_style"})
	end
	
	guif.textfield_drone_rename.text = drone.name
end

--------------------------------------------------------------------------------------
function switch_to_drone( player, player_mem, drone_new )
	-- character to drone
	local character = player.character
	
	if character == nil or (not character.valid) then
		player.print( {"gui-drones-wait-respawn"} )
		
	elseif character.name == "player" then
		debug_print( "man to drone..." )
		if drone_new.entity and drone_new.entity.valid then
			player.surface.create_entity({name = "flying-text", position = pos_offset(player.position,{x=-0.5,y=0.2}), text = {"gui-drones-flying-droning"}, color = {r=1, g=0, b=0}})
			
			player_mem.character = character
			--switch_entity_1to2( player, player.character, drone_new.entity )
			player.character = drone_new.entity -- go into drone
			drone_new.entity.active = true
			player.surface.create_entity({name = "flying-text", position = pos_offset(drone_new.entity.position,{x=-0.5,y=-1}), text = {"gui-drones-flying-link"}, color = {r=0, g=1, b=0}})
		end
		
		return( drone_new )
		
	elseif character.name == "drone-surveillance" or character.name == "drone-maintenance" then
		debug_print( "drone to drone..." )
		if drone_new.entity and drone_new.entity.valid then
			--switch_entity_1to2( player, player.character, drone_new.entity )
			player.character = drone_new.entity -- go into drone
			drone_new.entity.active = true
			player.surface.create_entity({name = "flying-text", position = pos_offset(drone_new.entity.position,{x=-0.5,y=-1}), text = {"gui-drones-flying-link"}, color = {r=0, g=1, b=0}})
		end
		
		return( drone_new )
	else
		player.print( {"gui-drones-not-in-player-drone"} )
	end
	
	return( nil )
end

--------------------------------------------------------------------------------------
function switch_to_character( player, fly )
	local player_mem = global.players_mem[player.index]
	local character = player_mem.character
	
	if character and character.valid and character.name == "player" then -- if body alive (death while droning detected in on_tick)
		debug_print( "back in body..." )
		--switch_entity_1to2( player, player.character, character )
		player.character = character -- reintegrate saved body
		if fly then
			player.surface.create_entity({name = "flying-text", position = pos_offset(player.position,{x=-0.5,y=0.2}), text = {"gui-drones-flying-back"}, color = {r=0, g=1, b=0}})
		end
		--player_mem.drone_cur = nil
		--player_mem.drone_dup = nil
	end
end

--------------------------------------------------------------------------------------
function switch_entity_1to2( player, ent1, ent2 )
	-- teleport ent1 to ent2
	local surface1 = ent1.surface
	local position1 = ent1.position
	local surface2 = ent2.surface
	local position2 = ent2.position
	player.teleport(ent2.position,ent2.surface)
	player.character = ent2
	ent1.teleport(position1, surface1) -- only for player between surface
end

--------------------------------------------------------------------------------------
function find_lost_character()
	local surf = game.surfaces["nauvis"]
	for chunk in surf.get_chunks() do
		a = {{chunk.x * 32, chunk.y * 32}, {chunk.x * 32 + 32, chunk.y * 32 + 32}}
		for k, entity in pairs(surf.find_entities_filtered({area=a, name= "player"})) do
			return entity
		end
	end
	return(nil)
end


--------------------------------------------------------------------------------------
local interface = {}

function interface.reset()
	debug_print( "reset" )
	
	local player
	for _, player in pairs(game.players) do
		global.players_mem[player.index] = nil
		init_player(player)
	end
	
	on_init()
end

function interface.lost()
	debug_print( "lost" )
	local lost = find_lost_character()
	
	if lost then
		local player, player_mem
		for _, player in pairs(game.players) do
			if player.connected then
				player_mem = global.players_mem[player.index]
				player.character = lost
				player_mem.drone_cur = nil
				player_mem.drone_cur_cover = -1
				update_menu_drones(player)
				break
			end
		end
	end
end

function interface.freedrones()
	debug_print( "freedrones" )
	
	local player
	for _, player in pairs(game.players) do
		player.insert({name = "drone-surveillance", count = 10} )
		player.insert({name = "drone-maintenance", count = 5} )
		player.insert({name = "drone-antenna", count = 5} )
	end
	
	on_init()
end

remote.add_interface( "drones", interface )

-- /c remote.call( "drones", "reset" )
-- /c remote.call( "drones", "lost" )
-- /c remote.call( "drones", "freedrones" )
